/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import ws.ladder.model.action.CollectionUpdateAction;
import ws.ladder.model.action.ElementAction;
import ws.ladder.model.action.Executable;
import ws.ladder.model.action.MigratorActionFactory;
import ws.ladder.model.action.PropertyKey;
import ws.ladder.model.action.SetterAction;
import ws.ladder.service.PersistenceOrchestrator;
import ws.ladder.util.ClassUtil;
import ws.ladder.util.ObjectCreator;

@Component
@Scope(value = "prototype")
public class DependencyGraph {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());
	public  static final String AGGREGATE_NAME_PREFIX = "AGGREGATE";

	private PersistenceOrchestrator persistenceOrchestrator;
	private ObjectCreator           objectCreator;
	private Map<String, Node>       aggregateRoots = new HashMap<String, Node>();	
	private Map<PropertyKey, List<Executable>> propertyActions = new HashMap<PropertyKey, List<Executable>>(); // Includes add, remove and reposition collection elements

	public DependencyGraph(ObjectCreator objectCreator) {
		this.objectCreator = objectCreator;
		this.persistenceOrchestrator = objectCreator.getDAS().getPersistenceOrchestrator();
	}

	private void getPotentialRoots(CallInfo ci) {
		// TODO: IN what order is this going to be saved?
		// Invoke save on all the non-dependent objects that have been modified
		int i = 1;
		for(ExtendedDataObject obj: objectCreator.getDataObjects()) {
			if(!obj.isReferenceType()) // Only reference types represent persistent entities
				continue;
			
			Node node = (Node) obj;
			// The aggregate root should either be the target root of this operation or be reachable from the target root 
			// Also the aggregate should either be transient or have been modified 
			if( (!node.isPersistent() || node.isAggregateModified()) && (!node.isDependent() && (node.getIncoming().size() > 0 || node == ci.getTarget())) )
				aggregateRoots.put(AGGREGATE_NAME_PREFIX+i++, node);
		}
	}
	
	private void findRealRoots() {
		Set<String> aggregateNames = new HashSet<String>(aggregateRoots.keySet());
		
		for(Map.Entry<String, Node> entry: aggregateRoots.entrySet()) {
			if(entry.getValue().getAggregateName() == null)
				entry.getValue().setAggregateSeal(entry.getKey());
		}
		
		// check to see which are not real aggregate roots
		for(String aggregateName: aggregateNames) {
			Node node = aggregateRoots.get(aggregateName);
			if(!node.getAggregateName().equals(aggregateName))
				aggregateRoots.remove(aggregateName);
		}
	}
	
	private void persistRoots() {
		// TODO: persist by topological sort 
		for(Node node: aggregateRoots.values())
			persistenceOrchestrator.saveOrUpdate(node.getInstance());
	}

	public void persistGraph(CallInfo callInfo) {
		getPotentialRoots(callInfo);
		findRealRoots();
		persistRoots();
	}

	public void processActions(Settings settings) {
		// Process the uni-directional actions
		Set<PropertyKey> uniDirKeys = new HashSet<PropertyKey>();
		for(PropertyKey key: propertyActions.keySet()) {
			if( !((ExtendedProperty)key.getProperty()).isBiDirectional() )
				uniDirKeys.add(key);
		}
		for(PropertyKey uniDirKey: uniDirKeys) {
			List<Executable> actions = propertyActions.remove(uniDirKey);
			processActions(uniDirKey, actions);
		}

		settings.getInterceptor().preBiDirActionStage( Collections.unmodifiableMap(propertyActions) );

		for(Map.Entry<PropertyKey, List<Executable>> entry: propertyActions.entrySet())
			processActions(entry.getKey(), propertyActions.get(entry.getKey()));
	}

	private void processActions(PropertyKey key, List<Executable> actions) {
		if(key.getProperty().isMany()) {
			for(Executable action: actions) {
				action.execute();
			}
		} else
			processToOne(actions);
	}

	public void processToOne(List<Executable> actions) {
		// We follow the system where we execute the action that has a non-null value if one is present or execute the null setting action
		// This policy is present to automatically fix broken links
		// The other policy we can have (maybe through configuration) is to throw an exception if a conflict is detected.
		// A conflict can occur if two actions set different values. It is still desirable to throw an exception if two different non-null values are being set.

		SetterAction actionToExecute = null;
		for(Executable action: actions) {
			SetterAction setterAction = (SetterAction) action;
			if(actionToExecute == null)
				actionToExecute = setterAction;
			if( ClassUtil.getInstance(setterAction.getValue()) != null) {
				if(ClassUtil.getInstance(actionToExecute.getValue()) != null && ClassUtil.getInstance(actionToExecute.getValue()) != ClassUtil.getInstance(setterAction.getValue()) )
					throw new IllegalArgumentException("Two different objects cannot share a reference to the same object in a ToOne bi-directional relationship.");
				else
					actionToExecute = setterAction;

			}
		}
		actionToExecute.execute();		
	}

	/**
	 * If this is a element action, the migrator action is found and the element action is recorded on it
	 * @param action
	 */
	public void addAction(Executable action) {

		List<Executable> actions = getActions(action.getKey());

		if (ElementAction.class.isAssignableFrom(action.getClass())) {
			CollectionUpdateAction collectionAction = getOrCreateMigratorAction(action.getKey());
			collectionAction.addTriggeredByOppositeAction((ElementAction) action);
		} else
			actions.add(action);
	}

	private List<Executable> getActions(PropertyKey key) {
		List<Executable> actions = propertyActions.get(key);
		if(actions == null) {
			actions = new ArrayList<Executable>();
			propertyActions.put((PropertyKey) key, actions);
		}		

		return actions;
	}

	public CollectionUpdateAction getOrCreateMigratorAction(PropertyKey key) {
		// actions should not be null
		List<Executable> actions = getActions(key);

		CollectionUpdateAction migrator = null;		
		if(actions.size() == 0) {
			migrator = MigratorActionFactory.getInstance(key);
			addAction(migrator);
			return migrator;
		}

		if(actions.size() == 1) {
			migrator = (CollectionUpdateAction) actions.get(0);
		} else
			throw new IllegalStateException("Collection property should have only one migrator action");

		return migrator;
	}

}