/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.model;

import java.util.List;
import java.util.Map;

import ws.ladder.exception.PropertyMappingException;
import ws.ladder.util.ObjectCreator;

public interface ExtendedDataObject extends DataObject {

	/**
	 * Return the ObjectCreator instance responsible for tracking the java object - DataObject instance mappings
	 * @return
	 */
	public ObjectCreator getObjectCreator();
	
	/**
	 * Get the actual object instance managed by this data object
	 */
	public Object getInstance();
	
	/**
	 * Set the reference to the actual java object
	 * @param instance
	 */
	public void setInstance(Object instance);
	
	/**
	 * Return a DataObject initialized according to the level specified in the parameters
	 */
	public DataObject read(Settings settings);
	
	/**
	 * Perform an efficient read using a query. This requires the use of a view
	 */
	public List<?> query(Settings settings);

	/**
	 * Update a persistent object according to the level specified in the parameters
	 */
	public ExtendedDataObject update(Settings settings);

	/**
	 * Create a data object from a data object and initialize according to the level specified in the
	 * parameters
	 * @see ExtendedDataObject#read
	 */
	public DataObject create(Settings settings);

	/**
	 * Copy the full data object instance based on the settings in the typeMapper
	 * @throws Exception 
	 */
	public DataObject clone(Settings settings);
	
	/**
	 * If this object represents a collection, then this function is used to get access to this information
	 * @return
	 */
	public List<ExtendedDataObject> getList();
	
	/**
	 * Get the property object based on the property path
	 */
	public Property getPropertyByPath(String path);	
	
	/**
	 * Return the path for this data object. 
	 * @return PATH_DELIMITER for the root data object and a non-null path for a contained data object
	 */
	public String getObjectPath();

	/**
	 * Data structure used to track the containment relationship within an object graph
	 * Useful for finding which object controls the life of another object
	 * 
	 * @return
	 */
	public DependencyGraph getDependencyGraph();	
	
	public boolean isVisited();
	
	public void setVisited(boolean visited);
	
	/**
	 * This method takes care of creating the containment graph and also any needed DataObject wrappers for the instance objects
	 */	
	public void createAggregate();

	/**
	 * returns true if the life of containedObj is controlled by this object
	 * @param object
	 * @return
	 */	
	boolean isAggregatePart(Object containedObj);

	boolean isRoot();

	public void actionOneToManySingle(CallInfo ci, AssociationAction actionType) throws PropertyMappingException;

	public void markLinkProcessed(String propertyName, ExtendedDataObject value );

	public boolean isLinkProcessed(String propertyName);

	public Object getLink(String propertyName);

	public void customRemoveFromCollection(CallInfo ci,
			ExtendedDataObject collection, ExtendedDataObject target);

	public void customAddToCollection(CallInfo ci,
			ExtendedDataObject collection, ExtendedDataObject target);

	void unlinkToOne(CallInfo ci) throws PropertyMappingException;

	void unlinkOneToMany(CallInfo ci, Object childrenToUnlink)
			throws PropertyMappingException;

	void setPersistent(boolean persistent);	
	
	/**
	 * Gets the id for a persistent instance
	 */
	public Object getIdentifierValue();

	void addEntity(ExtendedDataObject entity);

	ExtendedDataObject getEntity(ExtendedDataObject entity);

	void removeEntity(ExtendedDataObject entity);
	
	void checkInvariants();

	/**
	 * When used in a collection, what property is uniquely represented in a non-instance specific manner 
	 * This is set using an annotation on the domain model class. Used in path expressions.
	 * This is currently restricted to a single property.
	 */	
	public ExtendedProperty getCollectionKeyProperty();

	List<ExtendedDataObject> getList(CallInfo callInfo);

	void set(String propertyPath, Map<String, Object> propertyResult) throws Exception;
	
	/**
	 * Responsible for creating a new data object whose lifecycle is linked with this object.
	 * 
	 * @param id
	 * @param instanceType
	 * @param property
	 * @return
	 * @throws Exception
	 * @see {@link ExtendedDataObject#createDataObject(Object, ExtendedType)}
	 */
	public ExtendedDataObject createDataObject(Object id, Type instanceType, Property property) throws Exception;	

	/**
	 * This method is used to create a new data object that has to be in the scope of the data object that creates it.
	 * But its lifecycle is not linked with this object.
	 *
	 * @param id
	 * @param instanceType
	 * @return
	 * @throws Exception
	 * @see {@link ExtendedDataObject#createDataObject(Object, ExtendedType, Property)}
	 */
	public ExtendedDataObject createDataObject(Object id, Type instanceType) throws Exception;

	/**
	 * Used to set the container for a data object. This method is useful when the data object was created before its container.
	 * @param value
	 */
	public void setContainer(DataObject value);

	/**
	 * Set the containment property 
	 * @param value
	 */
	public void setContainmentProperty(Property value);

	boolean isReadOnly();

	void setReadOnly(boolean readOnly);

	/**
	 * This is not an embedded or a collection data object, but represents an entity data object
	 * @return
	 */
	public boolean isEntity();

	boolean isReferenceType();

	Type getReferenceType();

	ExtendedDataObject load(Settings settings);
}
