/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.model;

import java.util.List;

import javax.persistence.AccessType;

public interface ExtendedType extends BasicType {
	
	/**
	 * Should the access be based on property (getter method) or 
	 * by field.
	 * 
	 * @return
	 */
	public AccessType getAccessType();
	
	/**
	 * This is used to represent the Java class with meta data (e.g., JPA annotations) and business logic.
	 * 
	 * Types that are based off reference types are called derived types. They are based off a different Java class,
	 * but share the same properties and meta data with the reference type.
	 * @return true if reference type, false if derived type
	 */
	public boolean isReferenceType();
	
	/**
	 * For derived types, this will return the reference type from which it is based off.
	 * For reference types, this will return self.
	 * @return
	 */
	public ExtendedType getReferenceType();
	
	/**
	 * Returns the property that refers to the id of the entity
	 * @return
	 */
	public Property getIdentifierProperty();
	
	/**
	 * Returns the property that is marked as the user key
	 * @return
	 */
	public Property getUserKey();	
	
	/**
	 * Returns the property that is marked as the collection user key
	 * @return
	 */
	public Property getCollectionUserKey();	
	
	/**
	 * Returns the property that refers to the version of the entity
	 * @return
	 */
	public Property getVersionProperty();	
	
	/**
	 * Used to get access to the Reference type from derived types
	 * @return
	 */
	public TypeMapper getTypeMapper();	

	/**
	 * Get a list of all the types that are embedded in this type
	 * @return
	 */
	public List<Type> getEmbeddableTypes();
	
	/**
	 * Return true if the contents of this object cannot be changed
	 */
	public boolean isImmutable();
	
	/**
	 * Returns true if this type is an embedded type
	 */
	public boolean isEmbedded();
	
	/**
	 * Returns the highest ancestor entity in an inheritance hierarchy. Useful in EntityKey
	 */
	public ExtendedType getRootEntityType();
	
	/**
	 * Executes all the invariants associated with objects of this type and its ancestor types
	 */
	public void checkInvariants(Object object);

	/**
	 * Returns the name that the persistence provider uses to map an entity
	 * @see getName()
	 * @return
	 */
	public String getEntityName();

	/** 
	 * Returns true if this type represents an entity
	 * @return
	 */
	public boolean isEntity();
		
}
