/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.model;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

public class VOTypeMapper extends AbstractTypeMapper {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());
	
	public static final String VO_SUFFIX = "VO";

	private String referencePackagePath;
	private String derivedPackagePath;

	public String getReferencePackagePath() {
		return referencePackagePath;
	}

	public void setReferencePackagePath(String referencePackagePath) {
		this.referencePackagePath = referencePackagePath;
	}

	public String getDerivedPackagePath() {
		return derivedPackagePath;
	}

	public void setDerivedPackagePath(String derivedPackagePath) {
		this.derivedPackagePath = derivedPackagePath;
	}

	/* (non-Javadoc)
	 * @see org.LWSDO.model.TypeMapper#toReference(java.lang.Class)
	 */
	@Override
	public Class<?> toReference(Class<?> derivedClass) {
		if(isReference(derivedClass) || derivedClass.isPrimitive())
			return derivedClass;
		
		String fullClassName = derivedClass.getCanonicalName();
		String toClassName = fullClassName;

		if(fullClassName.startsWith(derivedPackagePath))
			// Change the package to the Reference package
			toClassName = fullClassName.replaceFirst(derivedPackagePath, referencePackagePath);  

		if(toClassName.endsWith(VO_SUFFIX) && toClassName.startsWith(referencePackagePath)) {
			// Remove the VO_SUFFIX
			int toNameLen = toClassName.length();
			StringBuffer toNameBuf = (new StringBuffer(toClassName)).delete(
					toNameLen - VO_SUFFIX.length(), toNameLen); 
			toClassName = toNameBuf.toString();		
		}

		try {
			return Class.forName(toClassName);
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}	

		return null;
	}

	/* (non-Javadoc)
	 * @see org.LWSDO.model.TypeMapper#toDerived(java.lang.Class)
	 */
	@Override
	public Class<?> toDerived(Class<?> referenceClass) {
		if(isDerived(referenceClass) || referenceClass.isPrimitive())
			return referenceClass;
		
		String fullClassName = referenceClass.getCanonicalName();
		String toClassName = fullClassName;

		if(fullClassName.startsWith(referencePackagePath))
			// Change the package to the Derived package
			toClassName = fullClassName.replaceFirst(referencePackagePath, derivedPackagePath);  

		if(!toClassName.endsWith(VO_SUFFIX) && toClassName.startsWith(derivedPackagePath))
			// Adjust the class name to the VO class name
			toClassName = toClassName.concat(VO_SUFFIX);                              

		try {
			return Class.forName(toClassName);
		} catch (ClassNotFoundException e) {
			logger.warn("Unable to find derived class: " + toClassName);
		}	

		return null;		
	}

	@Override
	public Class<?> getSourceClass(Class<?> clazz) {
		Class<?> result = null;

		switch(getDirection()) {
		case DERIVEDTOREFERENCE:
		case DERIVEDTODERIVED:			
			result = toDerived(clazz);
			break;
		case REFERENCETODERIVED:
		case REFERENCETOREFERENCE:			
			result = toReference(clazz);
			break;
		default:
			result = clazz;
			break;
		}

		return result;
	}
	
	@Override
	public Class<?> getTargetClass(Class<?> clazz) {
		Class<?> result = null;

		switch(getDirection()) {
		case DERIVEDTOREFERENCE:
		case REFERENCETOREFERENCE:			
			result = toReference(clazz);
			break;
		case REFERENCETODERIVED:
		case DERIVEDTODERIVED:			
			result = toDerived(clazz);
			break;
		default:
			result = clazz;
			break;
		}

		return result;
	}	
	
	@Override
	public boolean isDerived(Class<?> clazz) {
		return (clazz.getCanonicalName().startsWith(derivedPackagePath) && clazz.getCanonicalName().endsWith(VO_SUFFIX));
	}
	
	@Override
	public boolean isReference(Class<?> clazz) {
		return (clazz.getCanonicalName().startsWith(referencePackagePath) && !clazz.getCanonicalName().endsWith(VO_SUFFIX));
	}		

	@Override
	public TypeMapper newInstance(MapperDirection direction) {
		VOTypeMapper mapper = new VOTypeMapper();
		mapper.setDirection(direction);
		mapper.setDerivedPackagePath(getDerivedPackagePath());
		mapper.setReferencePackagePath(getReferencePackagePath());

		return mapper;		
	}	
}
