/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.model.view;

import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.model.CallInfo;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedType;
import ws.ladder.model.Settings;
import ws.ladder.service.QueryCapability;

/**
 * TODO: QueryBuilder needs to handle the following expressions

First Phase -- Restrict to what is needed by CSA 

CSA has the following filters
-----------------------------
ilike(param, %1)
in(param, %1)
idEq(%1)
ge(%1)
asc(%1)

e.g.

ilike(name, "%xyz%")
in("id", String[])
idEq(String)
ge("createdOn", long)
ge("updatedOn", long)
asc("name")

Modify view to support filters
------------------------------

ARTIFACTFILTER
id
name
displayName
description
iconUrl
detailedDescription
isCriticalSystemObject
ownedBy.name
state.name
artifactType.name
disabled
  FILTERS
  ilike(name, ":name")            // lower(name) LIKE :name
  in(state, ":state"              // name IN(:state)
  equal(ownedBy.name, ":owner")   // ownedBy.name = :owner 
  ge(createdOn, ":createdSince")  // createdOn > :createdSince
  ge(updatedOn, ":updatedSince"); // updaetdOn > :updatedSince
  asc(name)                       // ORDER BY name  also honor OrderBy annotation in model if no ORDER BY clause is provided

 *
 */
public abstract class QueryBuilder {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	private static final String SELECT_CLAUSE = "SELECT ";
	private static final String COMMA_DELIMITER = ", ";
	public  static final String ID_PARAMETER_NAME = "id_";

	private ExtendedDataObject entity;
	private ViewBranch view;
	private boolean SQL;
	private boolean addIdentifier;
	
	public ExtendedDataObject getEntity() {
		return entity;
	}	

	public boolean doAddIdentifier() {
		return addIdentifier;
	}

	public void setAddIdentifier(boolean addIdentifier) {
		this.addIdentifier = addIdentifier;
	}

	public abstract Query createQuery(CallInfo callInfo);

	public boolean isSQL() {
		return SQL;
	}

	public void init(ExtendedDataObject entity, ViewBranch view) {
		this.entity = entity;
		this.view = view;

		if(entity.getIdentifierValue() != null)
			addIdentifier = true;
	}

	private QueryCapability getQueryCapability()  {
		return entity.getObjectCreator().getDAS().getPersistenceOrchestrator().getQueryCapability();
	}

	public String constructQuery(Map<String, Object> filters) {
		//		First check for a SQL query 
		//		When retrieving the QueryBuilder instance, registered SQL queries are given preference over HQL/JPQL queries.
		//
		//		It is the responsibility of the user to ensure that the registered SQL is ANSI compliant 
		//		and can work with all the databases that the user uses.

		// TODO: handle sub-branches

		if(view.getContentView() != null && view.getContentView().getNativeQuery() != null) {
			SQL = true;
			String queryString = view.getContentView().getNativeQuery().getQueryString();
			if(doAddIdentifier()) {
				String idClause = view.getContentView().getNativeQuery().getIdentifierClause();
				if((idClause == null || idClause.trim().equals("")))
					throw new RuntimeException("Identifier clause expected in native query");
				
				queryString = queryString + " " + idClause + " = :" + ID_PARAMETER_NAME;
			}
			
			return queryString.replaceAll("[\n\r]", "");
		}

		StringBuffer HQL = new StringBuffer(constructHQL());
		HQL.append(buildWhereClause(filters));
		HQL.append(buildOrderClause());

		return HQL.toString();		
	}

	protected String constructHQL() {
		Map<String, ColumnMeta> meta = view.getAugmentedAttributes();

		StringBuffer HQL = new StringBuffer(SELECT_CLAUSE);

		int position = 0;
		for(ColumnMeta columnMeta: meta.values()) {
			if(!columnMeta.getViewProperty().doFetch()) // Should the property not be fetched
				continue;

			columnMeta.setPosition(position++);			
			if(!SELECT_CLAUSE.equals(HQL.toString()))
				HQL.append(COMMA_DELIMITER);
			String columnString = columnMeta.getQueryString(getQueryCapability());
			HQL.append(columnString);
		}

		HQL.append(" FROM " + entity.getReferenceType().getInstanceClass().getName() + " AS " + view.getViewProperty(ViewProperty.ROOT_PROPERTY_NAME).getAlias());

		// TODO: if more than one collection property is specified then we would get a severe performance hit due to cartesian product
		//       This can be solved by breaking the query intelligently. This is for the future, for now we shall log a warning
		// getCollectionBranches()

		for(ViewProperty viewProperty: view.getAliasedItems()) {
			ViewProperty parentView = viewProperty.getParent();
			if(parentView == null)
				continue; // the root needs to be skipped
			HQL.append(" LEFT OUTER JOIN " + viewProperty.getNormalizedName() + " AS " + viewProperty.getAlias());		
		}

		return HQL.toString();
	}	

	protected String buildOrderClause() {
		StringBuffer result = new StringBuffer();

		Map<Integer, String> orderClauses = new TreeMap<Integer, String>();
		for(ColumnMeta columnMeta: view.getAugmentedAttributes().values()) {
			if(columnMeta.getAttributePath().endsWith(ViewProperty.LIST_INDEX_ATTRIBUTE)) {
				String columnString = columnMeta.getQueryString(getQueryCapability());
				orderClauses.put(columnMeta.getAttributeLevel(), columnString);
			}
		}

		// filter order clauses
		StringBuffer filterOrderBy = new StringBuffer();
		for(Filter filter: view.getFilters()) {
			if(filter.isOrderBy()) {
				if(filterOrderBy.length() > 0)
					filterOrderBy.append(COMMA_DELIMITER);
				filterOrderBy.append(filter.getQueryString());
			}
		}

		// Append any ORDER BY clauses
		if(orderClauses.size() > 0 || filterOrderBy.length() > 0) {
			result.append(" ORDER BY ");
			StringBuffer orderString = new StringBuffer();
			for(String indexColumn: orderClauses.values())
				orderString.append( (orderString.length() > 0) ? COMMA_DELIMITER + indexColumn : indexColumn );

			if(filterOrderBy.length() > 0) 
				result.append((orderString.length() > 0) ? COMMA_DELIMITER + filterOrderBy : filterOrderBy);

			result.append(orderString);
		}		

		return result.toString();
	}

	protected String getNormalizedName(String propertyPath) {
		ViewProperty viewProperty = view.getViewProperty(ViewProperty.qualifyProperty(propertyPath));
		return viewProperty.getNormalizedName();
	}	

	protected String buildWhereClause(Map<String, Object> filters) {
		StringBuffer result = new StringBuffer();
		Map<String, Object> originalFilters = new HashMap<String, Object>(filters);
		filters.clear();

		Map<String, Parameter> parameterMap = new HashMap<String, Parameter>();
		for(Parameter parameter: view.getParameter())
			parameterMap.put(parameter.name, parameter);

		// Initialize the filterbypath map
		for(Filter filter: view.getFilters()) {
			if(filter.isOrderBy()) // Will be handled in the ORDER BY clause
				continue;

			Parameter param =  parameterMap.get(filter.getParameterName());
			if(!originalFilters.containsKey(Settings.encodeParam(filter.getParameterName())) ) { // parameter is not set
				// check if the parameter has a default value
				if(param != null)
					filters.put( Settings.encodeParam(filter.getParameterName()), filter.getNormalizedValue(param.defaultValue));
				else
					continue;
			} else {
				String key = Settings.encodeParam(filter.getParameterName());
				if(param != null)
					key = param.filterName;
				Object normalizedValue = filter.getNormalizedValue(originalFilters.get(Settings.encodeParam(key)));
				filters.put( Settings.encodeParam(filter.getParameterName()), normalizedValue);
			}
			addWhereStep(result);
		
			result.append(filter.getQueryString());
		}

		if(addIdentifier) {
			addWhereStep(result);

			result.append( getNormalizedName(((ExtendedType)entity.getReferenceType()).getIdentifierProperty().getName()) );
			result.append( " = :" + ID_PARAMETER_NAME);			
		}


		return result.toString();
	}	
	
	private void addWhereStep(StringBuffer result) {
		if(result.length() > 0)
			result.append(" AND ");		
		else
			result.append(" WHERE ");		
	}
}
