/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.model.view;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedDataObjectImpl;
import ws.ladder.model.ExtendedProperty;
import ws.ladder.model.ExtendedType;
import ws.ladder.model.Property;
import ws.ladder.model.Settings;
import ws.ladder.model.Type;
import ws.ladder.util.ClassUtil;

public class ViewBranch {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	private static final String ENTITY_ALIAS_PREFIX = "ENT";
	private static final String PROPERTY_ALIAS_PREFIX = "PROP";

	private static Map<ViewKey, ViewBranch>  viewCache = new ConcurrentHashMap<ViewKey, ViewBranch>();
	private        Type                      aggregateType;
	private        Map<String, ViewProperty> viewPropertyByPath = new LinkedHashMap<String, ViewProperty>();
	private        Map<String, ColumnMeta>   augmentedAttributes;
	private        Map<String, String>       attributes = new HashMap<String, String>();	
	private        String                    name;
	private        ContentView               contentView;
	private        boolean                   narrow;
	private        Set<Filter>               filters = new HashSet<Filter>();
	private        boolean                   crossAggregate;

	// Related to constructing child View Branches
	private        Set<ViewBranch>           subBranches;  // split according to parallel collections.
	private        ViewBranch                parentBranch; // branch containing the parent step
	private        ViewBranch                twig;         // attributes of simple type
	private        boolean                   collection;   // How many collections are under this branch

	private ViewBranch(ViewBranch parent, String rootName, Type rootType) {
		this.parentBranch = parent;
		this.name = rootName;
		this.aggregateType = rootType;
	}

	public ViewBranch(ViewKey viewKey, ContentView contentView) {

		this.contentView = contentView;
		this.aggregateType = viewKey.type;
		this.narrow = viewKey.narrow;
		setName(viewKey.viewName);

		// save the property alias before losing the order
		initAttributes(contentView.getAttributeList());

		init();

		if(contentView.getChildren() == null || contentView.getChildren().size() == 0) {
			// Perform automatic division into multiple queries if necessary
			createBranches();
			consolidateBranches();
			for(ViewBranch child: subBranches) {
				child.setAggregateType(aggregateType);
				child.init();
			}
		} else { // User has explicitly specified the queries needed to populate the view
			this.subBranches = new HashSet<ViewBranch>();
			for(ContentView child: contentView.getChildren()) {
				ViewBranch childViewBranch = new ViewBranch(this, null, this.aggregateType);
				childViewBranch.setContentView(child);
				childViewBranch.init();
				this.subBranches.add(childViewBranch);
			}
		}

		viewCache.put(viewKey, this);
		
		// Initialize filters
		normalizeFilters();
		for(ViewBranch child: subBranches)
			child.normalizeFilters();
	}

	private void initAttributes(List<String> attributeList) {
		if(attributes != null)
			attributes.clear();

		for(int i = 0; i < attributeList.size(); i++)
			addAttribute(attributeList.get(i));
	}

	private void init() {
		setFilters();
		createViewProperties();	
		setAliases();	
	}	
	
	public Set<Parameter> getParameter() {
		ContentView cView = this.contentView;
		if(cView == null && parentBranch != null)
			cView = parentBranch.getContentView();
		
		return (cView.getParameter() == null) ? new HashSet<Parameter>() : new HashSet<Parameter>(cView.getParameter());
	}

	private void setFilters() {
		ContentView cView = this.contentView;
		if(cView == null && parentBranch != null)
			cView = parentBranch.getContentView();
		
		if(cView == null || cView.getFilter() == null)
			return;
		
		// Copy the relevant filters from the content view
		for(Filter filter: cView.getFilter()) {
			Filter narrowedFilter = filter.narrow();
			filters.add(narrowedFilter);	
		}		
	}
	
	private void normalizeFilters() {
		// use the entity alias to replace the filter attribute names
		for(Filter filter: filters)
			filter.normalize(getNormalizedNames());
	}
	
	private Map<String, String> getNormalizedNames() {
		Map<String, String> result = new HashMap<String, String>();		
		
		for(String attribute: augmentedAttributes.keySet())
			result.put(attribute, viewPropertyByPath.get(attribute).getNormalizedName());

		return result;
	}

	public Set<Filter> getFilters() {
		return filters;
	}

	public void setFilters(Set<Filter> filters) {
		this.filters = filters;
	}	
	
	public ContentView getContentView() {
		return contentView;
	}

	public void setContentView(ContentView contentView) {
		this.contentView = contentView;
	}	

	public boolean hasCollection() {
		return collection;
	}

	public void setHasCollection(boolean collection) {
		this.collection = collection;
	}

	public Type getAggregateType() {
		return aggregateType;
	}

	public void setAggregateType(Type aggregateType) {
		this.aggregateType = aggregateType;
	}	

	public ViewBranch getParentBranch() {
		return parentBranch;
	}

	public void setParentBranch(ViewBranch parentBranch) {
		this.parentBranch = parentBranch;
	}	

	public boolean isCrossAggregate() {
		return crossAggregate;
	}

	public void setCrossAggregate(boolean crossAggregate) {
		this.crossAggregate = crossAggregate;
	}	

	/**
	 * Use the subBranches if there is more than one
	 * @return
	 */
	public Set<ViewBranch> getSubBranches() {
		return Collections.unmodifiableSet(this.subBranches);
	}

	public boolean hasParallelCollections() {
		return this.subBranches.size() > 1;
	}

	public final static class ViewKey {
		final Type type;
		final String viewName;
		final boolean narrow;

		public ViewKey(Type type, String viewName, boolean narrow) {
			this.type = type;
			this.viewName = viewName;
			this.narrow = narrow;
		}

		@Override
		public boolean equals(Object object) {
			if(!ViewKey.class.isAssignableFrom(object.getClass()))
				return false;

			ViewKey otherKey = (ViewKey) object;

			if(viewName.equals(otherKey.viewName) && 
					this.type == otherKey.type &&
					this.narrow == otherKey.narrow)
				return true;

			return false;
		}

		@Override
		public int hashCode() {
			int result = 17;
			result = 37 * result + viewName.hashCode();
			result = 37 * result + this.type.hashCode();
			result = 37 * result + ((narrow) ? 1 : 0);
			return result;
		}
	}

	private String getBranchPath() {
		StringBuffer result = new StringBuffer(this.getParentBranch() == null ? "" : name);
		ViewBranch parentViewBranch = getParentBranch();

		while(parentViewBranch != null) { 
			if(parentViewBranch.getParentBranch() == null)
				break;
			result.insert(0, getParentBranch().getName() + Settings.PATH_DELIMITER);
			parentViewBranch = parentViewBranch.getParentBranch();
		}

		return result.toString();
	}

	private String getBranchName(String attribute) {
		String branchPath = getBranchPath();
		if(attribute.length() <= branchPath.length())
			return null;

		String relativeName = (branchPath.length() > 0) ? attribute.substring(branchPath.length() + Settings.PATH_DELIMITER.length()) : attribute;
		return ViewProperty.getRootName(relativeName);
	}

	private void createBranches() {
		Map<String, ViewBranch> newBranches = new HashMap<String, ViewBranch>();

		for(String attribute: attributes.keySet()) {
			String branchName = getBranchName(attribute);
			Property rootProperty = aggregateType.getProperty(branchName);
			Type rootType = rootProperty.getType();
			if(rootProperty.isMany())
				rootType = ((ExtendedProperty)rootProperty).getElementType();			
			if(rootType.isDataType()) { // Do not create branches for simple types
				if(twig == null)
					twig = new ViewBranch(this, null, null);
				twig.addAttribute(attribute);
				continue;
			}

			ViewBranch branch = newBranches.get(branchName);
			if(branch == null) {
				// Create a new branch
				branch = new ViewBranch(this, branchName, rootType);
				if(rootProperty.isMany())
					branch.setHasCollection(true); // The number of collections in this branch is atleast 1 due to the root property.
				// There could be more collections due to the sub branches.
				newBranches.put(branchName, branch);
			}

			branch.addAttribute(attribute); // add actual attribute
		}

		// Recursively create branches
		for(ViewBranch branch: newBranches.values())
			branch.createBranches();

		this.subBranches = new HashSet<ViewBranch>(newBranches.values());
	}

	private void consolidateBranches() {
		// First go to the leaf
		for(ViewBranch subBranch: subBranches) 
			subBranch.consolidateBranches();

		if(this.subBranches.size() == 0)
			return; // nothing to consolidate

		// If a branch has only one child then consolidate by clearing its child and bringing any parallel branches to the top
		if(this.subBranches.size() == 1) {
			// Check if the only child has parallel collections. If so, the child has to be replaced with the parallel collections
			ViewBranch child = this.subBranches.iterator().next();

			if(child.subBranches.size() > 1) { // parallel collections
				subBranches.clear();

				// update the collection count before consolidation
				this.setHasCollection(this.collection || child.hasCollection());

				for(ViewBranch grandChild: child.subBranches) {
					// NOTE: Name is irrelevant in merge. So we don't do anything about updating it.
					grandChild.setAggregateType(child.getAggregateType());
					grandChild.setParentBranch(this);
					this.subBranches.add(grandChild); // make the grandChild the child
				}
			}
			child.subBranches.clear();

			return;
		}

		// If a branch has many children
		// First check if there are parallel collections
		Set<ViewBranch> parallelCollectionBranches = new HashSet<ViewBranch>();
		Set<ViewBranch> nonCollectionBranches = new HashSet<ViewBranch>(); // branches that need to be merged
		for(ViewBranch child: subBranches) {
			this.setHasCollection(this.collection || child.hasCollection()); // Mark if it has any collection in its child sub-branches
			if(!child.hasCollection())
				nonCollectionBranches.add(child);
			else
				parallelCollectionBranches.add(child);
		}

		if(parallelCollectionBranches.size() > 1) {
			Iterator<ViewBranch> iter = nonCollectionBranches.iterator();
			// need to consolidate the parallel branches
			while(iter.hasNext())
				for(ViewBranch parallelBranch: parallelCollectionBranches)
					parallelBranch.merge(nonCollectionBranches.iterator().next());		

			subBranches.clear();
			subBranches.addAll(parallelCollectionBranches);
		} else
			subBranches.clear(); // the current branch is sufficient

		for(ViewBranch child: subBranches)
			child.merge(twig);
	}

	private void merge(ViewBranch graftBranch) {
		Set<String> uniqueAttributes = new HashSet<String>(this.attributes.keySet());
		uniqueAttributes.addAll(graftBranch.attributes.keySet());
		this.initAttributes(new ArrayList<String>(uniqueAttributes));
	}

	private void addAttribute(String attribute) {
		if(!attributes.containsKey(attribute)) {
			attributes.put(attribute, PROPERTY_ALIAS_PREFIX + (attributes.size()+1));
		}
	}

	public static ViewBranch get(ViewKey viewKey, ContentView contentView) {
		if(viewCache.containsKey(viewKey))
			return viewCache.get(viewKey);

		ViewBranch result = new ViewBranch(viewKey, contentView);
		viewCache.put(viewKey, result);

		return result;
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}	

	private void setAliases() {
		int entityAliasCount = 1;
		for(ViewProperty eProp: viewPropertyByPath.values()) {
			if(eProp.getAlias() != null)
				throw new IllegalStateException("Alias has already been set");

			if(eProp.isEntity())
				eProp.setAlias(ENTITY_ALIAS_PREFIX + entityAliasCount++);
		}
	}

	protected String getParentPath(String propertyPath) {
		if(propertyPath.indexOf(Settings.PATH_DELIMITER) != -1)
			return propertyPath.substring(0, propertyPath.lastIndexOf(Settings.PATH_DELIMITER));
		else
			return ViewProperty.ROOT_PROPERTY_NAME;
	}	

	/**
	 * This method adds the anchor property and any additional properties needed to support the creation of the
	 * correct anchor object. This could be the identifier and/or the entity name if supported.
	 * 
	 * @param attribute
	 * @return
	 */
	private ViewProperty getParentViewProperty(String attribute, boolean doFetch) {
		attribute = ViewProperty.qualifyProperty(attribute);
		String parentPath = getParentPath(attribute);

		ViewProperty result = null;
		Property property = null;
		while(parentPath != null) {
			if(viewPropertyByPath.containsKey(parentPath)) {
				result = viewPropertyByPath.get(parentPath);
				property = result.getProperty();
			} else {
				property = aggregateType.getProperty(ViewProperty.unqualifyProperty(parentPath));
			}

			if(property == null && parentPath.equals(ViewProperty.ROOT_PROPERTY_NAME))
				break;

			// If this is a collection, then if the element is an entity type then break
			if(property.isMany()) {
				Type elementType = ((ExtendedProperty)property).getElementType();
				if(ExtendedType.class.isAssignableFrom(elementType.getClass())) {
					if(!((ExtendedType)elementType).isEmbedded() && !elementType.isDataType())
						break; // found it
				}
			} else { // if the entity type is not an embedddable or a simple type then break
				Type type = property.getType();
				if(ExtendedType.class.isAssignableFrom(type.getClass())) {
					if(!((ExtendedType)type).isEmbedded() && !type.isDataType())
						break; // found it
				}
			}
			parentPath = getParentPath(parentPath);
		}
		// If the anchor is not present, then add it
		if(result == null ) {
			result = new ViewProperty(parentPath, true, getParentViewProperty(parentPath, doFetch));
			viewPropertyByPath.put(ViewProperty.qualifyProperty(parentPath), result);
		} 
		result.setFetch(doFetch);

		return result;
	}

	private void createViewProperties() {
		ViewProperty viewProperty = new ViewProperty(true, aggregateType);
		viewPropertyByPath.put(ViewProperty.ROOT_PROPERTY_NAME, viewProperty);

		for(Map.Entry<String, String> attribute: attributes.entrySet())
			addViewProperty(attribute.getKey(), false, true, attribute.getValue());
		
		// Create view property objects for attributes in the filter list that are not in the select list
		for(Filter filter: filters) {
			if(!containsViewProperty(filter.getAttribute())) // should not be fetched
				addViewProperty(filter.getAttribute(), true, false, null);
		}
		
		setAugmentedAttributes();
	}
	
	private boolean containsViewProperty(String attribute) {
		if(viewPropertyByPath.containsKey(ViewProperty.qualifyProperty(attribute)))
			return true;
		
		return false;
	}
	
	private ViewProperty addViewProperty(String attribute, boolean isDynamic, boolean doFetch, String propertyAlias) {
	
		ViewProperty anchor = getParentViewProperty(attribute, doFetch);
		ViewProperty viewProperty = new ViewProperty(attribute, isDynamic, anchor);
		viewProperty.setPropertyAlias(propertyAlias);
		viewPropertyByPath.put(ViewProperty.qualifyProperty(attribute), viewProperty);
		viewProperty.setFetch(doFetch);
		
		return viewProperty;
	}

	public void setAugmentedAttributes() {
		if(narrow && (contentView == null || contentView.getNativeQuery() == null))
			throw new RuntimeException("Narrowing is only supported with native query. The persistence provider requires the whole object to be loaded inorder to infer the type, which we don't do.");

		Set<ColumnMeta> meta = new HashSet<ColumnMeta>();
		for(ViewProperty viewProperty: viewPropertyByPath.values())
			meta.addAll(viewProperty.getColumnMeta(narrow));

		// Sort the attributes
		List<String> sortedPath = new ArrayList<String>();
		for(ColumnMeta columnMeta: meta) 
			sortedPath.add(columnMeta.getAttributePath());
		Collections.sort(sortedPath);

		Map<String, ColumnMeta> augmentedAttributeMap = new HashMap<String, ColumnMeta>();
		for(ColumnMeta columnMeta: meta) {
			if(augmentedAttributeMap.containsKey(columnMeta.getAttributePath())) { // remove duplicates, preferring the property in the view
				if(!columnMeta.getViewProperty().isDynamic())
					augmentedAttributeMap.put(columnMeta.getAttributePath(), columnMeta);
			} else
				augmentedAttributeMap.put(columnMeta.getAttributePath(), columnMeta);
		}
		augmentedAttributes = new LinkedHashMap<String, ColumnMeta>(); // maintain the order
		for(String path: sortedPath) {
			ColumnMeta columnMeta = augmentedAttributeMap.get(path);
			augmentedAttributes.put(path, columnMeta);
			if(contentView != null && contentView.getNativeQuery() != null) { // check that the attribute is covered by the native query
				int position = contentView.getNativeQuery().getPosition(ViewProperty.unqualifyProperty(path));
				if(position == -1) {
					logger.warn("The native query does not populate the attribute " + ViewProperty.unqualifyProperty(path) + " for view: " + getViewName());
					contentView.getNativeQuery().setUsable(false);
				} else
					columnMeta.setPosition(position);
			}
		}
		
		// fix the content meta of dynamically added properties by the view property
		for(Map.Entry<String, ColumnMeta> augmented: augmentedAttributes.entrySet()) {
			if(viewPropertyByPath.get(augmented.getKey()) == null && !augmented.getValue().isDependent()) 
				augmented.getValue().setViewProperty(addViewProperty(augmented.getKey(), true, true, null));
		}
	}	

	private String getViewName() {
		if(parentBranch != null)
			return parentBranch.getName();
		else
			return getName();
	}

	public Map<String, ColumnMeta> getAugmentedAttributes() {
		return augmentedAttributes;
	}

	public ViewProperty getViewProperty(String propertyPath) {
		return viewPropertyByPath.get(propertyPath);
	}

	public List<ViewProperty> getAliasedItems() {
		List<String> aliasedPaths = new ArrayList<String>();
		List<ViewProperty> aliasedItems = new ArrayList<ViewProperty>();

		for(String path: viewPropertyByPath.keySet()) {
			ViewProperty viewProperty = viewPropertyByPath.get(path);
			if(viewProperty.getAlias() == null)
				continue;
			aliasedPaths.add(path);
		}
		Collections.sort(aliasedPaths);
		for(String aliasedPath: aliasedPaths)
			aliasedItems.add(viewPropertyByPath.get(aliasedPath));

		return aliasedItems;
	}

	public static String getNext(String propertyPath) {
		if(propertyPath.indexOf(Settings.PATH_DELIMITER) != -1)
			return propertyPath.substring(propertyPath.indexOf(Settings.PATH_DELIMITER)+1);
		else
			return null;
	}	

	public static String getTopAttribute(String propertyPath) {
		if(propertyPath == null || "".equals(propertyPath))
			return null;

		if(propertyPath.indexOf(Settings.PATH_DELIMITER) != -1) {
			return propertyPath.substring(0, propertyPath.indexOf(Settings.PATH_DELIMITER));
		} else
			return propertyPath;
	}

	public Object getQueryValue(Object[] queryResultRow, String path) {
		ColumnMeta columnMeta = augmentedAttributes.get(ViewProperty.qualifyProperty(path));
		if(columnMeta != null)
			return queryResultRow[columnMeta.getPosition()];
		else
			return null;
	}


	public ExtendedDataObject getQueryRoot(Object obj, ExtendedDataObject entity) throws Exception {
		ExtendedDataObject result = null;

		if(ClassUtil.getDimensionCount(obj) == 1) {
			Object[] queryRow = (Object[])obj;

			String idPropertyName = ((ExtendedType)entity.getType()).getIdentifierProperty().getName();
			Object idValue = getQueryValue(queryRow, idPropertyName);

			String entityName = (String) getQueryValue(queryRow, ViewProperty.ENTITYNAME_ATTRIBUTE);
			ExtendedType type = (ExtendedType) entity.getType();
			if(entityName != null) {
				type = (ExtendedType) entity.getObjectCreator().getDAS().getType(entityName);
				type = (ExtendedType) entity.getObjectCreator().getTargetType(type.getInstanceClass());
			}

			// find and create data object
			result = ((ExtendedDataObjectImpl)entity).getByEntityKey(idValue, entity.getType());
			if(result == null)
				result = entity.createDataObject(idValue, type);
		}
		if(result != null)
			result.setReadOnly(true);

		return result;
	}	

	public void normalize(ExtendedDataObject root, Object[] queryResultRow) throws Exception {
		Map<Integer, ColumnMeta> metaMap = new HashMap<Integer, ColumnMeta>();
		for(ColumnMeta columnMeta: augmentedAttributes.values())
			metaMap.put(columnMeta.getPosition(), columnMeta);

		Map<String, Object> propertyResult = new HashMap<String, Object>();
		for(int i = 0; i < queryResultRow.length; i++) {
			ColumnMeta columnMeta = metaMap.get(i);
			propertyResult.put(columnMeta.getAttributePath(), queryResultRow[i]);
		}

		for(String propertyPath: propertyResult.keySet()) {
			ColumnMeta meta = augmentedAttributes.get(propertyPath);
			if(meta.getViewProperty().isDynamic())
				continue;

			// Set the value and create any intermediate objects if necessary
			root.set(ViewProperty.unqualifyProperty(propertyPath), propertyResult);
		}
	}
}