/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.service;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.AutowireCapableBeanFactory;
import org.springframework.stereotype.Component;

import ws.ladder.model.AbstractType;
import ws.ladder.model.AggregateAction;
import ws.ladder.model.CallInfo;
import ws.ladder.model.DerivedType;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedProperty;
import ws.ladder.model.ExtendedType;
import ws.ladder.model.Property;
import ws.ladder.model.SimpleType;
import ws.ladder.model.Type;
import ws.ladder.model.TypeMapper;
import ws.ladder.model.view.QueryBuilder;

/**
 * Using IoC pattern
 * 
 * @author Dilip Dalton
 *
 */
@Component
public abstract class AbstractDataAccessService implements DataAccessService {

	@Autowired 
	protected AutowireCapableBeanFactory beanFactory;

	@Autowired
	protected SpringQueryBuilderFactory queryBuilderFactory;	

	protected PersistenceOrchestrator persistenceOrchestrator;
	protected TypeMapper typeMapper;	

	protected Map<String, Type> types = new ConcurrentHashMap<String, Type>();
	protected Map<String, Type> derivedTypes = new ConcurrentHashMap<String, Type>();	

	protected void addDerivedType(String className, Type type) {
		addType(className, type, derivedTypes);
	}	

	protected void addType(String className, Type type) {
		addType(className, type, types);
	}

	protected void addType(String className, Type type, Map<String, Type> typeMap) {
		typeMap.put(className, type);

		if(ExtendedType.class.isAssignableFrom(type.getClass())) {
			String entityName = ((ExtendedType)type).getEntityName();
			if(entityName != null && !className.equals(entityName)) {
				if(typeMap.containsKey(entityName)) {
					if(!className.equals(typeMap.get(entityName).getInstanceClass().getName()))
						throw new RuntimeException("Type " + typeMap.get(entityName).getName() + " already exists for entityName: " + entityName);
				} else
					typeMap.put(entityName, type);
			}
		}
	}

	@Override
	public Type getDerivedType(Class<?> clazz) {
		Type result = derivedTypes.get(clazz.getName());

		// create a Type object for this class
		if (result == null) {
			result = new SimpleType(clazz);
			addDerivedType(clazz.getName(), result);
		}

		return result;
	}	

	@Override
	public Type getType(String name) {
		return types.get(name);
	}

	@Override
	public Type getDerivedType(String name) {
		return derivedTypes.get(name);
	}	

	@Override
	public TypeMapper getTypeMapper() {
		return typeMapper;
	}

	@Override
	public PersistenceOrchestrator getPersistenceOrchestrator() {
		return persistenceOrchestrator;
	}

	@Override
	public void setPersistenceOrchestrator(PersistenceOrchestrator persistenceOrchestrator) {
		this.persistenceOrchestrator = persistenceOrchestrator;
	}		

	@Override
	public List<Type> getTypes() {
		return new ArrayList<Type>(types.values());
	}	

	/**
	 * This method uses the persistence mechanism to do any post-processing activity
	 * @param newInstance
	 * @return
	 */

	@Override
	public void postProcess(Object newInstance) {
		if(newInstance == null)
			return;

		// Since we are manually creating the instance we need to autowire any spring dependencies
		beanFactory.autowireBean(newInstance);		
	}

	private Object getByUserKey(CallInfo callInfo, ExtendedType type) {
		ExtendedDataObject from = (ExtendedDataObject) callInfo.getSource();
		Property userKeyProperty = type.getUserKey();

		if(userKeyProperty != null) {
			Map<String, String> param = new HashMap<String, String>();
			param.put(userKeyProperty.getName(), from.get(userKeyProperty).toString() );

			return getPersistenceOrchestrator().findByProperty(from.getReferenceType(), param);
		} else
			return null;
	}


	@Override
	public Object getPersistentObject(CallInfo callInfo, TypeMapper typeMapper) {
		ExtendedDataObject from = (ExtendedDataObject) callInfo.getSource();
		Object persistentObject = null;

		if(!ExtendedType.class.isAssignableFrom(from.getType().getClass()))
			return null;

		ExtendedType type = (ExtendedType) from.getType();

		if(type.isEmbedded()) // We don't separately load embedded values from the database
			return null;

		if(!(callInfo.getSettings().getAction() == AggregateAction.CLONE) )
			persistentObject = getByUserKey(callInfo, type);

		if(persistentObject == null) {
			ExtendedProperty identifierProperty = (ExtendedProperty) type.getIdentifierProperty();

			Serializable id = (Serializable) identifierProperty.getValue(from);
			if(id != null) {
				Class<?> desiredClass = typeMapper.toReference(type.getInstanceClass());
				persistentObject = getPersistenceOrchestrator().findById(desiredClass, id);
			} 
		}

		return persistentObject;
	}	

	protected void postProcess() {
		initDerived();

		initRootType();
	}

	private void initRootType() {
		for (Type type : types.values()) {
			if (AbstractType.class.isAssignableFrom(type.getClass())) {
				((AbstractType)type).initRootEntityType(this);
			}
		}
		for (Type type : derivedTypes.values()) {
			if (AbstractType.class.isAssignableFrom(type.getClass())) {
				((AbstractType)type).initRootEntityType(this);
			}
		}			
	}

	protected void initDerived() {
		for(Type type: types.values()) {
			if(SimpleType.class.isAssignableFrom(type.getClass()))
				continue;
			Class<?> derivedClass = typeMapper.toDerived(type.getInstanceClass());
			if(derivedClass != null) {
				Type derived = new DerivedType((ExtendedType) type, derivedClass);
				derivedTypes.put(derived.getName(), derived);
			}
		}

		// init the derived properties
		for (Type type : derivedTypes.values()) {
			if (DerivedType.class.isAssignableFrom(type.getClass())) {
				DerivedType derivedType = (DerivedType) type;
				derivedType.setProperty(this);
			}
		}		

		for (Type type : derivedTypes.values()) {
			if (DerivedType.class.isAssignableFrom(type.getClass())) {
				DerivedType derivedType = (DerivedType) type;
				derivedType.setOpposite(this);
			}
		}			
	}

	public QueryBuilder getQueryBuilder() {
		return queryBuilderFactory.create();
	}
}