/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.service;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.annotation.Resource;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hibernate.SessionFactory;
import org.hibernate.cfg.Configuration;
import org.hibernate.mapping.PersistentClass;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.orm.hibernate4.LocalSessionFactoryBean;

import ws.ladder.model.HibernateType;
import ws.ladder.model.SimpleType;
import ws.ladder.model.Type;
import ws.ladder.model.TypeMapper;

/**
 * This class is part of the Data Access Service framework
 * 
 * @author Dilip Dalton
 * 
 */
public class HibernateDAS extends AbstractDataAccessService {

	private static final Logger logger = LogManager.getLogger(new Exception()
	.getStackTrace()[0].getClassName());

	@Resource(name = "&sessionFactory")
	protected LocalSessionFactoryBean sessionFactoryBean;
	
	@Autowired
	protected SessionFactory sessionFactory;	
	
	public HibernateDAS(TypeMapper typeMapper) {
		this.typeMapper = typeMapper;
	}
	
	public SessionFactory getSessionFactory() {
		return sessionFactory;
	}

	@Override
	public Type getType(Class<?> clazz) {
		Type result = super.getType(clazz.getName());

		// create a Type object for this class
		if (result == null) {
			result = new SimpleType(clazz);
			addType(clazz.getName(), result);
		}

		return result;
	}

	public void define() {
		Configuration conf = sessionFactoryBean.getConfiguration();
		Iterator<PersistentClass> classMappings = conf.getClassMappings();

		logger.info("Getting the list of hibernate mapped classes");
		while (classMappings.hasNext()) {
			PersistentClass classMapping = (PersistentClass) classMappings
					.next();
			logger.debug("     Adding hibernate persisted class: "
					+ classMapping.getClassName());
			defineTypes(classMapping);
		}

		// Set the base types
		setBaseTypes();

		// Define the properties for the Types
		// This will end up defining the simple types
		defineProperties();
		
		postProcess();
	}

	protected void defineTypes(PersistentClass classMapping) {
		HibernateType dataType = new HibernateType(sessionFactory
				.getTypeHelper().entity(classMapping.getEntityName()),
				classMapping);
		logger.debug("Defined data type: " + dataType.getName());
		addType(dataType.getName(), dataType);
		
		for(Type type: dataType.getEmbeddableTypes()) {
			addType(type.getName(), type);
		}
	}

	protected void defineProperties() {
		for (Type type : types.values()) {
			if (HibernateType.class.isAssignableFrom(type.getClass())) {
				HibernateType hibernateType = (HibernateType) type;
				hibernateType.setProperty(this);
			}
		}

		// Link the bi-directional relationship between the properties
		for (Type type : types.values()) {
			if (HibernateType.class.isAssignableFrom(type.getClass())) {
				HibernateType hibernateType = (HibernateType) type;
				hibernateType.setOpposite(this);
			}
		}
	}

	protected void setBaseTypes() {
		for (Type type : types.values()) {
			if (HibernateType.class.isAssignableFrom(type.getClass())) {
				HibernateType hibernateType = (HibernateType) type;
				if(hibernateType.getHibernateType().isComponentType())
					continue;

				List<Type> baseTypes = new ArrayList<Type>();
				PersistentClass base = ((PersistentClass)hibernateType.getHibernateClass()).getSuperclass();

				if (base != null) {
					Type baseType = types.get(base.getEntityName());
					if (baseType != null)
						baseTypes.add(baseType);	
				}
				hibernateType.setBaseType(baseTypes);
			}
		}
	}

	@Override
	public List<String> getAggregateList() {
		List<String> result = new ArrayList<String>();
		
		Configuration conf = sessionFactoryBean.getConfiguration();
		Iterator<PersistentClass> classMappings = conf.getClassMappings();

		logger.info("Getting the list of hibernate mapped classes");
		while (classMappings.hasNext()) {
			PersistentClass classMapping = (PersistentClass) classMappings
					.next();
			result.add(classMapping.getEntityName());
		}
		
		return result;
	}
}