/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.util;

import java.beans.Introspector;
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.persistence.Access;
import javax.persistence.AccessType;

import net.sf.cglib.proxy.Enhancer;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hibernate.proxy.HibernateProxy;
import org.hibernate.proxy.LazyInitializer;
import org.springframework.aop.framework.Advised;
import org.springframework.aop.support.AopUtils;
import org.springframework.stereotype.Component;

import ws.ladder.annotation.CustomCollectionAdd;
import ws.ladder.annotation.CustomCollectionRemove;
import ws.ladder.annotation.CustomGetter;
import ws.ladder.annotation.CustomSetter;
import ws.ladder.annotation.DataModel;
import ws.ladder.annotation.Invariant;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.Settings;
import ws.ladder.model.Type;
import ws.ladder.model.event.CollectionEvent;
import ws.ladder.model.event.PropertyEvent;

@Component
public class ClassUtil {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());
	private static final String JAVASSIST_STARTWITH = "org.javassist.tmp.";
	private static final String JAVASSIST_INDEXOF = "_$$_javassist_";	

	private static Map<Class<?>, Map<String, Method>> getterMethods = new ConcurrentHashMap<Class<?>, Map<String, Method>>();
	private static Map<Class<?>, Map<String, Method>> setterMethods = new ConcurrentHashMap<Class<?>, Map<String, Method>>();	
	private static Map<Class<?>, Map<String, Field>> fields         = new ConcurrentHashMap<Class<?>, Map<String, Field>>();

	// related to annotation support for the framework
	private static Map<Class<?>, Map<String, Method>> customGetterMethods  = new ConcurrentHashMap<Class<?>, Map<String, Method>>();
	private static Map<Class<?>, Map<String, Method>> customSetterMethods  = new ConcurrentHashMap<Class<?>, Map<String, Method>>();
	private static Map<Class<?>, Map<String, Method>> customAddMethods     = new ConcurrentHashMap<Class<?>, Map<String, Method>>();
	private static Map<Class<?>, Map<String, Method>> customRemoveMethods  = new ConcurrentHashMap<Class<?>, Map<String, Method>>();
	private static Map<Class<?>, Set<Method>>         invariants           = new ConcurrentHashMap<Class<?>, Set<Method>>(); // includes invariant annotated methods of ancestors
	private static Map<Class<?>, Map<String, Annotation>> classAnnotations = new ConcurrentHashMap<Class<?>, Map<String, Annotation>>();

	private static final HashSet<Class<?>> WRAPPER_TYPES = getWrapperTypes();
	private static final HashSet<Class<?>> BASIC_TYPES = getBasicTypes();

	public static boolean isWrapperType(Class<?> clazz)
	{
		return WRAPPER_TYPES.contains(clazz);
	}

	private static HashSet<Class<?>> getWrapperTypes()
	{
		HashSet<Class<?>> ret = new HashSet<Class<?>>();
		ret.add(Boolean.class);
		ret.add(Character.class);
		ret.add(Byte.class);
		ret.add(Short.class);
		ret.add(Integer.class);
		ret.add(Long.class);
		ret.add(Float.class);
		ret.add(Double.class);
		ret.add(Void.class);
		return ret;
	}

	private static HashSet<Class<?>> getBasicTypes()
	{
		HashSet<Class<?>> ret = new HashSet<Class<?>>();
		ret.add(String.class);
		ret.add(java.math.BigDecimal.class);
		ret.add(java.sql.Date.class);
		ret.add(java.sql.Time.class);
		ret.add(java.sql.Timestamp.class);
		ret.add(java.util.Date.class);
		return ret;
	}    

	public static boolean isBasicType(Class<?> clazz) {
		return clazz.isPrimitive() || isWrapperType(clazz) || BASIC_TYPES.contains(clazz);
	}

	public Map<String, Field> getAllFields(Class<?> targetClass) {
		Map<String, Field> fieldObjs = fields.get(targetClass);
		if(fieldObjs == null)
			fieldObjs = collectFields(targetClass);

		Map<String, Field> result = new HashMap<String, Field>();

		// process inherited classes
		while (targetClass != Object.class) {
			fieldObjs = fields.get(targetClass);
			if(fieldObjs == null)
				fieldObjs = collectFields(targetClass);			
			result.putAll(fieldObjs);

			// climb to the super class and repeat
			targetClass = targetClass.getSuperclass();			
		}

		return result;
	}

	protected static void getCustomAnnotations(Class<?> beanClass) {	
		synchronized(beanClass) {
			invariants.put(beanClass, new HashSet<Method>());		
			customGetterMethods.put(beanClass, new HashMap<String, Method>());
			customSetterMethods.put(beanClass, new HashMap<String, Method>());
			customAddMethods.put(beanClass, new HashMap<String, Method>());
			customRemoveMethods.put(beanClass, new HashMap<String, Method>());		

			for(Method method: beanClass.getMethods()) {
				if(method.getAnnotation(Invariant.class) != null) {
					if(method.getParameterTypes().length != 0 && !PropertyEvent.class.isAssignableFrom(method.getParameterTypes()[0]))
						throw new RuntimeException("An invariant method should have a single argument and that argument should implement PropertyEvent");
					invariants.get(beanClass).add(method);
				}

				if(method.getAnnotation(CustomGetter.class) != null) {
					CustomGetter customGetter = method.getAnnotation(CustomGetter.class);
					if(method.getParameterTypes().length != 0 && !PropertyEvent.class.isAssignableFrom(method.getParameterTypes()[0]))
						throw new RuntimeException("A custom getter method should have a single argument and that argument should implement PropertyEvent");
					customGetterMethods.get(beanClass).put(customGetter.property(), method);
				}

				if(method.getAnnotation(CustomSetter.class) != null) {
					CustomSetter customSetter = method.getAnnotation(CustomSetter.class);
					if (method.getParameterTypes().length != 1 && !PropertyEvent.class.isAssignableFrom(method.getParameterTypes()[0]))
						throw new RuntimeException("A custom setter method should have a single argument and that argument should implement PropertyEvent");
					customSetterMethods.get(beanClass).put(customSetter.property(), method);
				}

				if(method.getAnnotation(CustomCollectionAdd.class) != null) {
					CustomCollectionAdd customCollectionAdd = method.getAnnotation(CustomCollectionAdd.class);
					if (method.getParameterTypes().length != 1 && !CollectionEvent.class.isAssignableFrom(method.getParameterTypes()[0]))
						throw new RuntimeException("A custom collectionAdd method should have a single argument and that argument should implement CollectionEvent");				
					customAddMethods.get(beanClass).put(customCollectionAdd.property(), method);
				}

				if(method.getAnnotation(CustomCollectionRemove.class) != null) {
					CustomCollectionRemove customCollectionRemove = method.getAnnotation(CustomCollectionRemove.class);
					if (method.getParameterTypes().length != 1 && !CollectionEvent.class.isAssignableFrom(method.getParameterTypes()[0]))
						throw new RuntimeException("A custom collectionRemove method should have a single argument and that argument should implement CollectionEvent");				
					customRemoveMethods.get(beanClass).put(customCollectionRemove.property(), method);
				}			
			}
		}
	}	

	public static Set<Method> getInvariants(Class<?> targetClass) {
		Set<Method> methodObjs = invariants.get(targetClass);		
		if(methodObjs == null) {
			getCustomAnnotations(targetClass);
			methodObjs = invariants.get(targetClass);
		}

		return methodObjs;
	}


	public static Method getCustomGetter(Class<?> targetClass, String targetProperty) {
		Map<String, Method> methodObjs = customGetterMethods.get(targetClass);		
		if(methodObjs == null) {
			getCustomAnnotations(targetClass);
			methodObjs = customGetterMethods.get(targetClass);
		}

		return methodObjs.get(targetProperty);
	}

	public static Method getCustomSetter(Class<?> targetClass, String targetProperty) {
		Map<String, Method> methodObjs = customSetterMethods.get(targetClass);		
		if(methodObjs == null) {
			getCustomAnnotations(targetClass);
			methodObjs = customSetterMethods.get(targetClass);
		}

		return methodObjs.get(targetProperty);
	}	

	public static Method getCustomAdd(Class<?> targetClass, String targetProperty) {		
		Map<String, Method> methodObjs = customAddMethods.get(targetClass);		
		if(methodObjs == null) {
			getCustomAnnotations(targetClass);
			methodObjs = customAddMethods.get(targetClass);
		}

		return methodObjs.get(targetProperty);
	}	

	public static Method getCustomRemove(Class<?> targetClass, String targetProperty) {	
		Map<String, Method> methodObjs = customRemoveMethods.get(targetClass);		
		if(methodObjs == null) {
			getCustomAnnotations(targetClass);
			methodObjs = customRemoveMethods.get(targetClass);
		}

		return methodObjs.get(targetProperty);
	}		

	public static Map<String, Field> collectFields(Class<?> targetClass) {

		// Get all fields declared by the class or interface.
		// This includes public, protected, default (package) access, 
		// and private methods, but excludes inherited methods.
		Map<String, Field> map = new HashMap<String, Field>();

		while (targetClass != Object.class) {
			Field[] fa = targetClass.getDeclaredFields();

			for (int i=fa.length-1; i > -1; i--) {
				Field f = fa[i];

				final int mod = f.getModifiers();
				f.setAccessible( true );

				if (Modifier.isStatic(mod))
					continue;

				map.put(f.getName(), f); 
			}
			// climb to the super class and repeat
			targetClass = targetClass.getSuperclass();
		}
		fields.put(targetClass, map);

		return map;
	}	

	public static Field getField(Class<?> targetClass, String targetProperty){
		Map<String, Field> fieldObjs = fields.get(targetClass);
		if(fieldObjs == null)
			fieldObjs = collectFields(targetClass);

		return fieldObjs.get(targetProperty);
	}	

	protected static Map<String, Method> getAllGetterMethods(Class<?> beanClass) {
		// Get all methods declared by the class or interface.
		// This includes public, protected, default (package) access, 
		// and private methods, but excludes inherited methods.
		Map<String, Method> map = new HashMap<String, Method>();

		while (beanClass != Object.class) {
			Method[] ma = beanClass.getDeclaredMethods();

			for (int i=ma.length-1; i > -1; i--) {
				Method m = ma[i];

				if (!m.getName().startsWith("get") && !m.getName().startsWith("is"))
					continue;
				if (m.getParameterTypes().length != 0)
					continue;
				final int mod = m.getModifiers();

				if (Modifier.isStatic(mod))
					continue;

				// Adds the specified element to the set if it is not already present
				final String propertyString = getGetterProperty(m);
				String propertyName = Introspector.decapitalize(propertyString);
				if(map.get(propertyName) == null || (map.get(propertyName) != null && map.get(propertyName).getName().startsWith("is"))) // give preference to the "get" method
					map.put(propertyName, m);
			}
			// climb to the super class and repeat
			beanClass = beanClass.getSuperclass();
		}
		getterMethods.put(beanClass, map);

		return map;
	}	

	public static String getGetterProperty(Method m) {
		return (m.getName().startsWith("get")) ? m.getName().substring("get".length()) : m.getName().substring("is".length());
	}	

	public static Method getGetterMethod(Class<?> targetClass, String targetProperty){
		Map<String, Method> methodObjs = getterMethods.get(targetClass);
		if(methodObjs == null)
			methodObjs = getAllGetterMethods(targetClass);

		return methodObjs.get(targetProperty);
	}

	protected static Map<String, Method> getAllSetterMethods(Class<?> beanClass) {

		// Get all methods declared by the class or interface.
		// This includes public, protected, default (package) access, 
		// and private methods, but excludes inherited methods.
		Map<String, Method> map = new HashMap<String, Method>();

		while (beanClass != Object.class) {
			Method[] ma = beanClass.getDeclaredMethods();

			for (int i=ma.length-1; i > -1; i--) {
				Method m = ma[i];

				if (!m.getName().startsWith("set"))
					continue;
				if (m.getParameterTypes().length != 1)
					continue;
				final int mod = m.getModifiers();

				if (Modifier.isStatic(mod))
					continue;

				// Adds the specified element to the set if it is not already present
				final String propertyString = m.getName().substring("set".length());
				String propertyName = Introspector.decapitalize(propertyString);
				map.put(propertyName, m);
			}
			// climb to the super class and repeat
			beanClass = beanClass.getSuperclass();
		}
		setterMethods.put(beanClass, map);

		return map;
	}		

	public static Method getSetterMethod(Class<?> targetClass, String targetProperty){
		Map<String, Method> methodObjs = setterMethods.get(targetClass);
		if(methodObjs == null)
			methodObjs = getAllSetterMethods(targetClass);

		return methodObjs.get(targetProperty);
	}	

	public static String getMappedBy(Class<?> targetClass, String propertyName) {
		String result = null;

		Method getterMethod = getGetterMethod(targetClass, propertyName);
		Field field         = getField(targetClass, propertyName);

		javax.persistence.OneToOne oneToOne = getterMethod.getAnnotation(javax.persistence.OneToOne.class);
		javax.persistence.OneToMany oneToMany = getterMethod.getAnnotation(javax.persistence.OneToMany.class);
		javax.persistence.ManyToMany manyToMany = getterMethod.getAnnotation(javax.persistence.ManyToMany.class);
		if(oneToOne == null && oneToMany == null && manyToMany == null) {
			oneToOne = field.getAnnotation(javax.persistence.OneToOne.class);
			oneToMany = field.getAnnotation(javax.persistence.OneToMany.class);
			manyToMany = field.getAnnotation(javax.persistence.ManyToMany.class);
		}

		if(oneToOne != null && oneToOne.mappedBy() != null)
			result = oneToOne.mappedBy();		
		else if(oneToMany != null && oneToMany.mappedBy() != null)
			result = oneToMany.mappedBy();
		else if(manyToMany != null && manyToMany.mappedBy() != null)
			result = manyToMany.mappedBy();

		return result;
	}

	public static boolean isUpdatable(Class<?> jPAClass, String propertyName) {
		boolean result = true;

		Method getterMethod = getGetterMethod(jPAClass, propertyName);
		Field field         = getField(jPAClass, propertyName);

		javax.persistence.Column column = getterMethod.getAnnotation(javax.persistence.Column.class);
		if(column == null)
			column = field.getAnnotation(javax.persistence.Column.class);

		if(column != null && !column.updatable())
			result = false;

		return result;
	}

	private static boolean isJavassistEnhanced(Class<?> c) {
		String className = c.getName();

		// pattern found in javassist ProxyFactory
		return className.startsWith(JAVASSIST_STARTWITH)
				|| className.indexOf(JAVASSIST_INDEXOF) != -1;
	}	

	public static boolean isEnhanced(Class<?> c) {
		boolean result = false;
		if (Enhancer.isEnhanced(c) || isJavassistEnhanced(c)) {
			result = true;
		}

		return result;
	}

	public static boolean isEnhanced(Object proxy) {
		boolean result = false;
		if (Enhancer.isEnhanced(proxy.getClass()) || isJavassistEnhanced(proxy.getClass()) ||
				AopUtils.isJdkDynamicProxy(proxy)) {
			result = true;
		}

		return result;
	}	

	public static Object getTargetObject(Object object) {
		Object result = object;

		// Hibernate based implementation, safe to initialize the proxy since we need its data
		// Allow user to override this behavior
		if (object instanceof HibernateProxy) {
			while(object instanceof HibernateProxy) {
				HibernateProxy proxy = (HibernateProxy) object; 
				LazyInitializer li = proxy.getHibernateLazyInitializer();
				object = li.getImplementation();
			}		
			return object;
		} else if (AopUtils.isJdkDynamicProxy(object)) {
			try {
				return ((Advised)object).getTargetSource().getTarget();
			} catch(Exception e) {
				throw wrapRun(e);
			}
		}

		return result;
	}

	/** 
	 * Invoke the given method as a privileged action, if necessary. 
	 * @throws InvocationTargetException 
	 * @throws IllegalAccessException 
	 * @throws IllegalArgumentException 
	 */
	public static Object invokeMethodAsPrivileged(final Object target, final Method method, final Object[] args) 
			throws InvocationTargetException, IllegalAccessException 
			{
		if (Modifier.isPublic(method.getModifiers()))
			return method.invoke(target, args);
		return AccessController.doPrivileged(
				new PrivilegedAction<Object>() {
					public Object run() {
						method.setAccessible(true);
						Object result = null;
						try {
							result = method.invoke(target, args);
						} catch (Exception e) {
							throw wrapRun(e);
						}
						return result;
					}
				});
			}

	/** 
	 * Invoke the given method as a privileged action, if necessary. 
	 * @throws InvocationTargetException 
	 * @throws IllegalAccessException 
	 * @throws IllegalArgumentException 
	 */
	public static Object invokeFieldAsPrivileged(final Object target, final Field field, final Object value, final boolean read) 
			throws InvocationTargetException, IllegalAccessException 
			{

		if (Modifier.isPublic(field.getModifiers())) {
			Object readValue = null;
			if(read)
				readValue = field.get(target);
			else
				field.set(target, value);
			return readValue;
		} else {
			return AccessController.doPrivileged(
					new PrivilegedAction<Object>() {
						public Object run() {
							field.setAccessible(true);
							Object readValue = null;
							try {
								if(read)
									readValue = field.get(target);
								else
									field.set(target, value);
							} catch (Exception e) {
								throw wrapRun(e);
							} 
							return readValue;
						}
					});
		}
			}		

	public static RuntimeException wrapRun(Exception e) {
		if(InvocationTargetException.class.isAssignableFrom(e.getClass())) {
			InvocationTargetException ite = (InvocationTargetException) e;
			Throwable cause = ite.getCause();
			if(cause != null && Exception.class.isAssignableFrom(cause.getClass()))
				e = (Exception) cause;
		}

		if(RuntimeException.class.isAssignableFrom(e.getClass()))
			return (RuntimeException) e;
		else
			return new RuntimeException(e);
	}

	public static AccessType getAccessStrategy(org.hibernate.cfg.AccessType type) {
		if ( org.hibernate.cfg.AccessType.PROPERTY.equals( type ) ) {
			return javax.persistence.AccessType.PROPERTY;
		}
		else if ( org.hibernate.cfg.AccessType.FIELD.equals( type ) ) {
			return javax.persistence.AccessType.FIELD;
		}
		else {
			return javax.persistence.AccessType.PROPERTY;
		}
	}	

	public static AccessType getAccessStrategy(String accessType) {
		return AccessType.valueOf(accessType);
	}	

	public static AccessType getHibernateAccessType(Class<?> instanceClass) {
		org.hibernate.cfg.AccessType classDefinedAccessType;

		org.hibernate.cfg.AccessType hibernateDefinedAccessType = org.hibernate.cfg.AccessType.DEFAULT;
		org.hibernate.cfg.AccessType jpaDefinedAccessType = org.hibernate.cfg.AccessType.DEFAULT;

		org.hibernate.annotations.AccessType accessTypeAnno = instanceClass.getAnnotation( org.hibernate.annotations.AccessType.class );
		if ( accessTypeAnno != null ) {
			hibernateDefinedAccessType = org.hibernate.cfg.AccessType.getAccessStrategy( accessTypeAnno.value() );
		}

		Access accessAnno = instanceClass.getAnnotation( Access.class );
		if ( accessAnno != null ) {
			jpaDefinedAccessType = org.hibernate.cfg.AccessType.getAccessStrategy( accessAnno.value() );
		}

		if ( hibernateDefinedAccessType != org.hibernate.cfg.AccessType.DEFAULT
				&& jpaDefinedAccessType != org.hibernate.cfg.AccessType.DEFAULT
				&& hibernateDefinedAccessType != jpaDefinedAccessType ) {
			throw new RuntimeException(
					"@AccessType and @Access specified with contradicting values. Use of @Access only is recommended. "
					);
		}

		if ( hibernateDefinedAccessType != org.hibernate.cfg.AccessType.DEFAULT ) {
			classDefinedAccessType = hibernateDefinedAccessType;
		}
		else {
			classDefinedAccessType = jpaDefinedAccessType;
		}
		return ClassUtil.getAccessStrategy(classDefinedAccessType);		
	}	

	public static int getDimensionCount(Object array) {
		int count = 0;
		Class<?> arrayClass = array.getClass();
		while ( arrayClass.isArray() ) {
			count++;
			arrayClass = arrayClass.getComponentType();
		}
		return count;
	}	

	public static Object getInstance(Object dataObject) {
		if(dataObject == null)
			return null;

		Object instance = dataObject;
		if(ExtendedDataObject.class.isAssignableFrom(dataObject.getClass()))
			instance = ((ExtendedDataObject)dataObject).getInstance();

		return instance;
	}	

	protected static Map<String, Annotation> getClassAnnotations(Class<?> beanClass) {

		// Get all annotations declared by the class.
		// This includes public, protected, default (package) access, 
		// and private methods, but excludes inherited methods.
		Map<String, Annotation> map = new HashMap<String, Annotation>();

		while (beanClass != Object.class) {
			Annotation[] annotations = beanClass.getAnnotations();
			for (int i=annotations.length-1; i > -1; i--) {
				Annotation a = annotations[i];
				map.put(a.annotationType().getName(), a);
			}

			// climb to the super class and repeat
			beanClass = beanClass.getSuperclass();
		}
		classAnnotations.put(beanClass, map);

		return map;
	}		

	public static Annotation getClassAnnotation(Class<?> targetClass, Class<?> annotationClass) {
		Map<String, Annotation> annotationObjects = classAnnotations.get(targetClass);
		if(annotationObjects == null)
			annotationObjects = getClassAnnotations(targetClass);	

		return annotationObjects.get(annotationClass.getName());
	}		

	public static String getCollectionUserKey(Type type) {

		Annotation annotation = getClassAnnotation(type.getInstanceClass(), DataModel.class);
		if(annotation != null && annotation.annotationType() == DataModel.class) {
			if( ((DataModel)annotation).collectionUserKey() == true)
				return ((DataModel)annotation).userKeyProperty();	
		}

		return null;
	}	

	public static String getUserKey(Type type) {

		Annotation annotation = getClassAnnotation(type.getInstanceClass(), DataModel.class);
		if(annotation != null && annotation.annotationType() == DataModel.class) {
			return ((DataModel)annotation).userKeyProperty();	
		}

		return null;
	}		

	public static String getBucketName(Class<?> clazz) {
		String name = clazz.getName();
		return Settings.encodeParam(name);
	}
}